% Deklarativ Programozas 2004 tavaszi felev
% Nagy hazi feladat keretprogram
% Drtos Mrton, Szeredi Pter, 2004.04.01, 1.4 valtozat

:- module(felhok_keret,
	  [
	   helyes_feladvany/1,  % helyes_feladvany(radarkep::in)
	   felhok_be/2,		% felhok_be(file::in, radarkep::out)
	   felhok_ki/3,		% felhok_ki(file::in, radarkep::in,
	                        %          felhok::in)
	   megold/2,		% megold(file::in, file::in)
	   stopper/2		% stopper(file::in, file::in)
	  ]
	 ).

:- use_module(library(lists), [member/2,append/3,select/3,same_length/2]).
:- meta_predicate forall(:, :).

%-------------- Tipusok ------------------------------------

% :- type radarkep        ---> f(list(sorosszeg), list(oszloposszeg), list(ref_pont)).
% :- type ref_pont        ---> f(sor, oszlop) | e(sor, oszlop).
% :- type sor               == integer.
% :- type oszlop            == integer.
% :- type sorosszeg         == integer.
% :- type oszloposszeg      == integer.
% :- type felhok            == list(felho).
% :- type felho           ---> f(kezdosor, kezdooszlop, zarosor, zarooszlop).
% :- type kezdosor          == sor.
% :- type kezdooszlop       == oszlop.
% :- type zarosor           == sor.
% :- type zarooszlop        == oszlop.

% :- pred felhok(radarkep::in, felhok::out).


%-------------- Bejaratok ------------------------------------

% felhok_be(+File, -Feladvany):
% A File-bol beolvashato a benne megadott Feladvany.
felhok_be(File, Feladvany) :-
	open_file(File, read, Stream),
	call_cleanup(read_in(Stream, Characters),
		     close(Stream)),
	parse_data(Characters, Feladvany).

% felhok_ki(+File, +Feladvany, +Felhok):
% A Feladvany Felhok megoldasat kiirja a File allomanyba.
felhok_ki(File, Feladvany, Felhok) :-
	open_file(File, write, Stream),
	current_output(OOut),
	set_output(Stream),
	call_cleanup(pretty_print(Feladvany, Felhok),
		     (set_output(OOut),close(Stream))).

% megold(+FileIn, +FileOut):
% A FileIn allomanybol beolvasott feladat minden megoldasat
% kiirja a FileOut allomanyba.
megold(FileIn, FileOut) :-
	felhok_be(FileIn, Radarkep),
	ensure_loaded(user:felhok),
	open_file(FileOut, write, Stream),
	current_output(OOut),
	set_output(Stream),
	call_cleanup(print_solutions(Radarkep, _),
		     (set_output(OOut),close(Stream))).

% stopper(+FileIn, +FileOut):
% Mint megold/2, de a futashoz szukseges idot is kiirja.
stopper(FileIn, FileOut) :-
	ensure_loaded(user:felhok),
	statistics(runtime, [T0|_]),
	megold(FileIn, FileOut),
	statistics(runtime, [T1|_]),
	T is T1 - T0,
	format('Feladat: ~a, futsi id: ~3d s~n', [FileIn,T]).

helyes_feladvany(f(SorOsszegek, OszlopOsszegek, RefPL)) :-
        length(SorOsszegek, N),
        length(OszlopOsszegek, M),
        % a referenciapontok koordintinak KoordL listja
        % az eredetivel azonos sorrend, ahol minden
        % referenciapont eget vagy felht jell:
        koordinatak(RefPL, KoordL),
        % az sszegek legkisebb rtke -1 lehet,
        % de nem haladhatjk meg a tbla mrett:
        forall( member(S, SorOsszegek), (-1 =< S, S =< M) ),
        forall( member(O, OszlopOsszegek), (-1 =< O, O =< N) ),
        % a KoordL lista minden eleme helyes:
        forall( member(Koord, KoordL), helyes_koordinata(Koord, N-M) ),
        % a KoordL lista lexikografikusan rendezett;
        % ebbl kvetkezen RefPL is:
        sort(KoordL, KoordL),
	% A KoordL lista brmely kt eleme klnbz;
        % ebbl kvetkezen a RefPL listban szerepl
	% (Sor,Oszlop) koordintk is pronknt klnbznek
        % (a rendezettsg miatt elegend ezt a szomszdos elemekre kiktni):
        forall( append(_, [K1, K2|_], KoordL), K1 \= K2 ).

%-------------- Ellenorzes -----------------------------------

helyes_koordinata(S-O, MaxS-MaxO) :-
        integer(S), S > 0, S =< MaxS,
        integer(O), O > 0, O =< MaxO.

%-------------- Seged-eljarasok ------------------------------

koordinatak([], []).
koordinatak([RefP|RefPL], [S-O|KoordL]) :-
	pont_koordinatak(RefP, S, O),
        koordinatak(RefPL, KoordL).

pont_koordinatak(e(S,O), S, O).
pont_koordinatak(f(S,O), S, O).

%  forall(Cl1, Cl2) : Cl1 minden megoldsa teljesti Cl2-t.
forall(Goal1, Goal2) :-
        (   Goal1, \+ Goal2 -> fail
        ;   true
        ).

% warning(+Txt, +Args):
% Kiirja a Txt szoveggel es Args argumentumlistaval
% megadott figyelmezteto szoveget.
warning(Txt, Args) :-
	print_message(warning, format(Txt, Args)).

% between(+I, +J, ?X): X egesz, amelyre I =< X, X =< J
between(I, J, I) :- I =< J.
between(I, J, X) :-
	I < J, I1 is I+1,
	between(I1, J, X).

%-------------- Allomanyok kezelese --------------------------

% open_file(+File, +Mode, -Stream):
% A File allomanyt megnyitja a Stream folyam formajaban Mode modban.
open_file(user, M, S) :-
	!, standard_stream(M, S).
open_file(File, Mode, Stream) :-
	open(File, Mode, Stream).

% standard_stream(?Mode, ?Stream):
% A Mode modu szabvanyos folyam a Stream.
standard_stream(read, user_input).
standard_stream(write, user_output).

%-------------- Beolvasas ------------------------------------

% read_in(+S, -L):
% S folyambol beolvashato karakterek listaja L.
read_in(S, L) :-
	get_code(S, C),
	read_in(C, S, L).

% read_in(+C, +S, -L):
% Ha C az S folyambol elozoleg beolvasott karakter, akkor L az S-bol
% beolvashato maradek karakterek listaja, C-t is beleertve.
% Ha C = -1, akkor L = [].
read_in(C, _, L) :-
	C =:= -1, !, L = [].
read_in(C, S, [C|L]) :-
	get_code(S, C1),
	read_in(C1, S, L).

% parse_data(+Chars, -Table):
% Chars fuzerbol kielemezheto egy Feladvany feladvnyler
parse_data(Chars, Feladvany) :-
	descriptor(Feladvany, Chars, Rest),
	correct_syntax(Rest),
	helyes_feladvany(Feladvany), !.
parse_data(_, _) :-
        warning('A bemeneti llomny nem megfelel formj.',[]),
	fail.

% correct_syntax(+Rest):
% A Rest fuzer a bemenet helyes maradeka, azaz ures.
correct_syntax(Rest) :-
	Rest = [_|_], !,
	chars_of_type(mid_line, L, Rest, _),
	warning('Nem vrt karakter(ek): "~s"', [L]),
	fail.
correct_syntax(_).

%-------------- DCG elemzes ----------------------------------

% Kielemezheto egy feladvany-leiro
descriptor(f(SOL,OOL,RefPL)) -->
	int_number_list(SOL),
	newline,
	int_number_list(OOL),
	newline,
	triplet_list(RefPL),
	chars_of_type(layout, _).

% Kielemezheto egy (tizes szamrendszerbeli) egesz szamokbol allo lista.
int_number_list([N|Ns]) -->
        int_number(N),
        int_number_list(Ns).
int_number_list([]) --> [].

% Kielemezheto egy (tizes szamrendszerbeli) Int egesz szam, amit
% esetleg szokozok eloznek meg.
int_number(Int) -->
	chars_of_type(horizontal, _),
        chars_of_type(digit, Ds),
	{Ds = [_|DTail],
	 \+ member(0'-, DTail),
	 number_codes(Int, Ds)}.

% triplet_list(RefPs)
% Kielemezheto egy harmasokbol allo RefPs lista.
triplet_list([RefP|RefPL]) -->
	hint_type(H),
        int_number(S), int_number(O),
	newline, !,
	{RefP =.. [H,S,O]},
	triplet_list(RefPL).
triplet_list([]) --> [].

% Kielemezheto egy referenciapont-leiro tipusa, amit
% esetleg szokozok eloznek meg.
hint_type(H) -->
        chars_of_type(horizontal, _),
        chars_of_type(hint, Hs),
        {Hs = [_], atom_codes(H, Hs)}.

newline -->
	chars_of_type(horizontal, _),
	char_of_type(vertical, _).

% layout_char(Dir, C):
% horizontal_layout_char(C): C  egy vizszintes iranyu
% nem lathato karakter.
horizontal_layout_char(0' ).
horizontal_layout_char(0'\t).
horizontal_layout_char(0'\r).

% vertical_layout_char(C): C  egy fuggoleges iranyu
% nem lathato karakter.
vertical_layout_char(0'\n).
vertical_layout_char(0'\f).
vertical_layout_char(0'\v).

% Kielemezheto egy Type tipusu C karakter.
char_of_type(Type, C) -->
	[C], {char_type(Type, C)}.

% chars_of_type(Type, Cs) -->
% Kielemezheto Type tipusu karakterek egy Cs listaja.
chars_of_type(Type, [C|Cs]) -->
	char_of_type(Type, C), !,
	chars_of_type(Type, Cs).
chars_of_type(_, []) --> [].

% char_type(Type, C) :
% A C karakter Type tipusu.
char_type(layout, C) :-
	(   vertical_layout_char(C) -> true
	;   horizontal_layout_char(C)
	).
char_type(digit, D) :-
	(   D >= 0'0, D =< 0'9 -> true
        ;   D == 0'-
        ).
char_type(mid_line, C) :-
	\+ vertical_layout_char(C).
char_type(horizontal, C) :-
	horizontal_layout_char(C).
char_type(vertical, C) :-
	vertical_layout_char(C).
char_type(hint, C) :-
	(   C == 0'f -> true
	;   C == 0'e
	).

%-------------- Kiiratas -------------------------------------

% print_solutions(+Descr, -Sols):
% A Descr altal megadott feladat osszes megoldasa Sols, ezeket
% szep formatumban kiirja az aktualis kimenetre.
print_solutions(Descr, Sols) :-
	catch(sorted_solutions(Descr, Sols),
	      repeated_clouds(Sol),
	      print_error(Sol)), !,
	(   member(Sol, Sols),
	    pretty_print(Descr, Sol), nl,
	    fail
	;   true
	).

% Hibajelzest ad, mert a Sol megoldasban ismetlodo felhok vannak.
% Mindig meghiusul.
print_error(Sol) :-
	format('A megoldsok nem rhatk ki.~n', []),
	format('Ismtld felhk vannak egy megoldson bell:~n~w~n', [Sol]),
	fail.

% sorted_solutions(+Descr, -Sols):
% A Descr altal megadott feladat osszes megoldasa, rendezett sorrendben, Sols.
sorted_solutions(Descr, Sols) :-
	findall(Clouds-1, sorted_clouds(Descr, Clouds), Sols1),
	keysort(Sols1, SortedSols1), % keysort nem tvoltja el a dupliktumokat.
	findall(Clouds, member(Clouds-_, SortedSols1), Sols).

% A Descr feladvny megoldsnak rendezett alakja Clouds.
% (Ahhoz, hogy a megoldasokat rendezni tudjuk, mindegyik megoldast (felholistat)
% valamilyen szabvanyos alakra kell hozni. Ezt a legegyszerubben rendezessel
% erhetjuk el.)
sorted_clouds(Descr, Clouds) :-
	user:felhok(Descr, Clouds0),
	sort(Clouds0, Clouds),
	(   same_length(Clouds0, Clouds) -> true
	;   raise_exception(repeated_clouds(Clouds0))
	).
	

% pretty_print(Descr)
% Kiirja a Descr tablat a felhok nelkul.
pretty_print(Descr) :-
	pretty_print(Descr, []).

% pretty_print(Descr, Clouds)
% Kiirja a Descr tabla megoldasait Clouds felhok eseten
pretty_print(Descr, Clouds) :-
	Descr = f(SOL,OOL,_),
	length(OOL, W),
        print_separator(W),
        coords(Clouds, [], Ccoords),
	bagof(Point:Values,
	      bagof(Value,
		    table_elem(Descr, Ccoords, Point, Value),
		    Values),
	      Elems0),
	findall(Point:Value, point_value(Point, Value, Elems0), Elems1),
	sort(Elems1, Elems),
	print_lines(SOL, 1, W, Elems),
        print_sum_line(OOL).

% point_value(Point, Value, Elems)
% A Point ponthoz az Elems lista altal rendelt megjelenitendo ertek a Value. 
% Az Elems lista Point:Values alaku elemekbol all. Ha a Values lista egyelemu,
% akkor Value ez az egyetlen elem lesz. Ha viszont Values tobbelemu, akkor
% Value = '!' lesz, ezzel hivva fel a figyelmet az inkonzisztenciara.
point_value(Point, Value, Elems) :-
	member(Point:Values, Elems),
	(   Values = [Value] -> true
	;   Value = '!'
	).

% table_elem(Descr, Clouds, Point, Value):
% A Descr feladvany Clouds megoldasanak megfelelo tablan
% a Point koordinataju pontban megjelenitendo a Value ertek.
table_elem(f(_, _, RefPL), Clouds, S-O, Value) :-
        (   member(f(S,O), RefPL), Value = '+'
	;   member(e(S,O), RefPL), Value = '-'
	;   member(S-O, Clouds),
	    \+ member(f(S,O), RefPL),
	    Value = '#'
	).

% print_lines(R0, C, SOL, Elems):
% Kiirja a C oszlopbol allo tabla R0-adik sorat a sorosszeggel
% egyutt, az Elems tartalomlistanak megfeleloen.
% A tartalomlista (Sor-Oszlop):Ertek alaku elemekbol all,
% es azt irja elo, hogy a (Sor,Oszlop) pozicion megjelenitendo
% az Ertek konstans. SOL a sorosszegek listaja.
print_lines([], _, _, _).
print_lines([SO|SOL], R0, C, Elems0) :-
	print_line(1, C, R0, SO, Elems0, Elems1),
	print_separator(C),
	R1 is R0+1,
	print_lines(SOL, R1, C, Elems1).

% print_line(C0, C, R0, SO, Elems0, Elems):
% Kiirja a C oszlopbol allo tabla R0-adik soranak a C0 oszlopban
% kezdodo reszet az Elems0 tartalomlistanak megfeleloen, SO
% sorosszeggel egyutt. Az Elems argumentumban visszaadja a
% tartalomlistanak az R0 sor utani reszet.
print_line(C0, C, _, SO, Elems, Elems) :-
	C0 > C, !,
	format('| ~d~n', [SO]).
print_line(C0, C, R0, SO, Elems0, Elems) :-
	field(R0-C0, Value, Elems0, Elems1),
	format('| ~w ', [Value]),
	C1 is C0+1,
	print_line(C1, C, R0, SO, Elems1, Elems).

% field(Point, Value, Elems0, Elems):
% A Point koordinataju pontban, az Elems0 tartalomlista szerint
% megjelenitendo ertek a Value. Elems a Point utani resznek megfelelo
% tartalomlista.
field(Point, Value, [Point:Value|Elems], Elems) :- !.
field(_, '.', Elems, Elems).

% print_separator(N):
% Kiir egy N hossz elvalaszto sort.
print_separator(0) :-
	!, write('+'), nl.
print_separator(N) :-
	write('+---'),
	N1 is N-1,
	print_separator(N1).

% coords(CloudL, CoordsL0, CoordsL) :
% A CloudL listaban szereplo felhok altal lefedett koordinatak
% halmazanak es a CoordsL0-ban szereplo koordinataknak az unioja
% CoordsL.
coords([], CoordsL, CoordsL).
coords([Cloud|CloudL], CoordsL0, CoordsL) :-
        findall(S-O, in_cloud(Cloud, S, O), CoordsL1, CoordsL0),
        coords(CloudL, CoordsL1, CoordsL).

% in_cloud(Cloud, S, O):
% A Cloud felho egy pontja S-O.
in_cloud(Cloud, S, O) :-
	Cloud = f(KS,KO,ZS,ZO),
	between(KS, ZS, S),
	between(KO, ZO, O).

% print_sum_line(OOL) :
% Kiirja az OOL oszloposszeglista elemeit tablazatszeruen,
% figyelembe veve a szamjegyek szamat.
print_sum_line([]) :-
	nl.
print_sum_line([X|XL]) :-
	number_codes(X, Cs),
	length(Cs, Len),
	After is 1//Len,
	Before is 4-Len-After,
        format('~*c~d~*c', [Before, 0' ,X, After, 0' ]),
        print_sum_line(XL).


